{ Provides a Question dialog with a variable number of buttons, each with user-supplied caption
  and modal result value. Button icons are added automatically corresponding to the button's
  modal result. The dialog function returns the modal result of the clicked button.
  The dialog message can optionally be emboldened.
  A button caption can optionally be followed by 'isDefault' or 'isCancel' to specify that button
  as the Default (pressed via Enter) or Cancel (pressed via Esc) button.

  function QuestionDlgIcons(const aCaption, aMsg: String; aDlgID: TMsgDlgType;
                            aButtons: array of const; isBoldMsg: Boolean=False): TModalResult;

  The format of the buttons array must be a repeated sequence of
    [ (modal result value), ('button caption'), (optional default/cancel string), ...]
  e. g.
    [mrNo,'No',  mrYes,'Yes','isDefault',  mrClose,'Close','isCancel']

  Copyright (C) 2018 H Page-Clark

  This library is free software; you can redistribute it and/or modify it under the terms of the
  GNU Library General Public License as published by the Free Software Foundation; either version
  2 of the License, or (at your option) any later version with the following modification:

  As a special exception, the copyright holders of this library give you permission to link this
  library with independent modules to produce an executable, regardless of the license terms of
  these independent modules,and to copy and distribute the resulting executable under terms of
  your choice, provided that you also meet, for each linked independent module, the terms and
  conditions of the license of that module. An independent module is a module which is not derived
  from or based on this library. If you modify this library, you may extend this exception to your
  version of the library, but you are not obligated to do so. If you do not wish to do so, delete
  this exception statement from your version.

  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
  the GNU Library General Public License for more details.

  You should have received a copy of the GNU Library General Public License along with this
  library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street - Fifth Floor,
  Boston, MA 02110-1335, USA.
}

unit AskDialog;

{$mode objfpc}{$H+}
{$ModeSwitch advancedrecords}

interface

uses
  Classes, SysUtils, Types, Dialogs, LCLType, UITypes, Forms, Buttons, Graphics,
  ExtCtrls, Controls;

type
  TBtnIDArray = array[mrNone..mrLast] of Integer;

  TDlgIDArray = array[TMsgDlgType] of Integer;

const
  BtnIDFromModalResult: TBtnIDArray = (idButtonBase, idButtonOk, idButtonCancel, idButtonAbort,
                                       idButtonRetry, idButtonIgnore, idButtonYes, idButtonNo,
                                       idButtonAll, idButtonNoToAll, idButtonYesToAll, idButtonClose);

  DlgIDFromMsgDlgType: TDlgIDArray = (idDialogWarning, idDialogError, idDialogInfo, idDialogConfirm, idDialogShield);

  Spacing = 15;

type

  { TButtonRec }

  TButtonRec = record
    caption: String;
    mr: TModalResult;
    isCancel: Boolean;
    isDefault: Boolean;
    procedure Init(const aCaption: String; aMR: TModalResult; anIsCancel, anIsDefault: Boolean);
  end;
  TButtonRecArray = array of TButtonRec;

  { TButtonRecList }

  TButtonRecList = record
  private
    function GetCount: Integer;
  public
    arr: TButtonRecArray;
    function AddButton(const aCaption: String; aMR: TModalResult; anIsCancel, anIsDefault: Boolean): Integer;
    function HasCancel(out btnIndex: Integer): Boolean;
    function HasDefault(out btnIndex: Integer): Boolean;
    procedure Init;
    property Count: Integer read GetCount;
  end;

   { TDlgIconCaptionButton }

   TDlgIconCaptionButton = class(TCustomBitBtn)
   public
     constructor Create(TheOwner: TComponent; const aCaption: String; aMR: TModalResult); reintroduce;
   end;
   TDlgIconCaptionButtonArray = array of TDlgIconCaptionButton;

   { TIconCaptionRecord }

   TIconCaptionRecord = record
   private
     FBitmap: TCustomBitmap;
     FCaption: String;
     FIconRect: TRect;
     FPerimeter: TRect;
     FSpacing: Integer;
     FTextPoint: TPoint;
   public
     procedure Init(const aBmp: TCustomBitmap; const aCaption: String; const aCnv: TCanvas; aSpacing: Integer);
     property Bitmap: TCustomBitmap read FBitmap;
     property Caption: String read FCaption;
     property IconRect: TRect read FIconRect;
     property Perimeter: TRect read FPerimeter;
     property TextPoint: TPoint read FTextPoint;
   end;

  { TQuestionDlg }

  TQuestionDlg = class(TCustomForm)
  private
    FButtonArray: TDlgIconCaptionButtonArray;
    FButtonPanel: TPanel;
    FCancelIndex: Integer;
    FDefaultIndex: Integer;
    FIconCaptionRec: TIconCaptionRecord;
    FIsBoldMessage: Boolean;
    procedure LayoutDialog;
  protected
    procedure DoShow; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure Paint; override;
  public
    constructor Create(const aCaption, aMsg: String; aDlgID: TMsgDlgType; aButtonList: TButtonRecList;
                       isBoldMsg: Boolean); reintroduce;
    destructor Destroy; override;
  end;

function QuestionDlgIcons(const aCaption, aMsg: String; aDlgID: TMsgDlgType; aButtons: array of const;
                          isBoldMsg: Boolean=False): TModalResult;

implementation

function ButtonsArrToButtonList(aButtons: array of const): TButtonRecList;

  function GetNextText(var index: Integer; out phrase: String): Boolean;
  begin
    phrase := '';
    if (index > High(aButtons)) then
      Exit(False);
    Result := True;
    case aButtons[index].VType of
      vtString:     phrase := aButtons[index].VString^;
      vtAnsiString: phrase := AnsiString(aButtons[index].VAnsiString);
      vtChar:       phrase := aButtons[index].VChar;
      vtPChar:      phrase := aButtons[index].VPChar;
      vtPWideChar:  phrase := aButtons[index].VPWideChar;
      vtWideChar:   phrase := AnsiString(aButtons[index].VWideChar);
      vtWidestring: phrase := AnsiString(WideString(aButtons[index].VWideString));
      else Result := False;
    end;
    if Result then
      Inc(index);
  end;

var
  mr: TModalResult;
  caption, txt, lowerTxt: String;
  default, cancel: Boolean;
  i: Integer;
begin
  Result.Init;
  i := Low(aButtons);
   while i <= High(aButtons) do
     begin
       if aButtons[i].VType <> vtInteger then
         raise Exception.Create('QuestionDlgIcons: integer expected at ' + IntToStr(i) +
                                ' but VType='+IntToStr(Ord(aButtons[i].VType))+' found.');
       if aButtons[i].VType = vtInteger then
         begin
           mr := aButtons[i].VInteger;
           Inc(i);

           txt := '';
           caption := '';
           default := False;
           cancel := False;

           while GetNextText(i, txt) do
             begin
               lowerTxt := LowerCase(txt);
               case lowerTxt of
                 'isdefault': begin
                                if caption = '' then
                                  raise Exception.Create('QuestionDlgIcons: missing button caption');
                                default := True;
                              end;
                 'iscancel': begin
                                if caption = '' then
                                  raise Exception.Create('QuestionDlgIcons: missing button caption');
                                cancel := True;
                              end;
                 else caption := txt;
               end;
             end;
         end;
       Result.AddButton(caption, mr, cancel, default);
     end;
end;

function QuestionDlgIcons(const aCaption, aMsg: String; aDlgID: TMsgDlgType;
  aButtons: array of const; isBoldMsg: Boolean): TModalResult;
var
  dlg: TQuestionDlg;
  btnList: TButtonRecList;
begin
  btnList := ButtonsArrToButtonList(aButtons);
  dlg := TQuestionDlg.Create(aCaption, aMsg, aDlgID, btnList, isBoldMsg);
  try
    Result := dlg.ShowModal;
  finally
    dlg.Free;
  end;
end;

{ TIconCaptionRecord }

procedure TIconCaptionRecord.Init(const aBmp: TCustomBitmap; const aCaption: String; const aCnv: TCanvas; aSpacing: Integer);
var
  sz: TSize;
begin
  FBitmap := aBmp;
  FCaption := aCaption;
  FSpacing := aSpacing;
  FIconRect := Rect(FSpacing, FSpacing, FSpacing+FBitmap.Width, FSpacing+FBitmap.Height);
  sz := aCnv.TextExtent(FCaption);
  FPerimeter := Rect(0, 0, sz.cx + FSpacing*3 + FBitmap.Width, (FSpacing shl 1) + FBitmap.Height);
  FTextPoint := Point((FSpacing shl 1) + FBitmap.Width, (FBitmap.Height div 2) + FSpacing - (sz.cy div 2));
end;

{ TDlgIconCaptionButton }

constructor TDlgIconCaptionButton.Create(TheOwner: TComponent; const aCaption: String; aMR: TModalResult);
begin
  inherited Create(TheOwner);
  Caption := aCaption;
  LoadGlyphFromStock(BtnIDFromModalResult[aMR]);
  ModalResult := aMR;
  AutoSize := True;
end;

{ TQuestionDlg }

procedure TQuestionDlg.LayoutDialog;
var
  btnsW: Integer = 0;
  panelW, panelH, topW, topH, btnSpacing: Integer;
  btn: TDlgIconCaptionButton;
begin
  for btn in FButtonArray do
    Inc(btnsW, btn.Width);
  panelW := btnsW + Spacing*Succ(Length(FButtonArray));
  topW := FIconCaptionRec.Perimeter.Width;
  panelH := FButtonArray[0].Height + Spacing shl 1;
  topH := FIconCaptionRec.Bitmap.Height + Spacing shl 1;
  case topW >= panelW of
    True:  begin
             btnSpacing := (topW - btnsW - Spacing shl 1) div Pred(Length(FButtonArray));
             FButtonPanel.ChildSizing.HorizontalSpacing := btnSpacing;
             FButtonPanel.SetBounds(Spacing, topH + Spacing, topW, panelH);
             SetBounds(Left, Top, topW, topH + panelH);
           end;
    False: begin
             FButtonPanel.SetBounds(Spacing, topH + Spacing, panelW, panelH);
             SetBounds(Left, Top, panelW, topH + panelH);
           end;
  end;
end;

procedure TQuestionDlg.DoShow;
begin
  LayoutDialog;
end;

procedure TQuestionDlg.KeyDown(var Key: Word; Shift: TShiftState);
begin
  case Key of
    VK_ESCAPE: if (FCancelIndex > -1) then begin
                 Key := 0;
                 FButtonArray[FCancelIndex].Click ;
               end;
    VK_RETURN: if (FDefaultIndex > -1) then begin
                 Key := 0;
                 FButtonArray[FDefaultIndex].Click;
               end;
    else inherited KeyDown(Key, Shift);
  end;
end;

procedure TQuestionDlg.Paint;
begin
  Canvas.StretchDraw(FIconCaptionRec.IconRect, FIconCaptionRec.Bitmap);
  if FIsBoldMessage then
    Canvas.Font.Style := [fsBold];
  Canvas.TextOut(FIconCaptionRec.TextPoint.X, FIconCaptionRec.TextPoint.Y, FIconCaptionRec.Caption);
end;

constructor TQuestionDlg.Create(const aCaption, aMsg: String;
  aDlgID: TMsgDlgType; aButtonList: TButtonRecList; isBoldMsg: Boolean);
var
  i: Integer;
begin
  inherited CreateNew(Nil);
  BorderStyle := bsDialog;
  Position:=poScreenCenter;
  KeyPreview := True;
  Caption := aCaption;
  Color := clBtnFace;
  aButtonList.HasCancel(FCancelIndex);
  aButtonList.HasDefault(FDefaultIndex);
  FIconCaptionRec.Init(GetDialogIcon(DlgIDFromMsgDlgType[aDlgID]), aMsg, Canvas, Spacing);
  FIsBoldMessage := isBoldMsg;

  FButtonPanel := TPanel.Create(Self);
  with FButtonPanel do begin
    BevelOuter := bvNone;
    ChildSizing.ControlsPerLine := aButtonList.Count;
    ChildSizing.Layout := cclLeftToRightThenTopToBottom;
    ChildSizing.HorizontalSpacing := Spacing;
    Top := FIconCaptionRec.Perimeter.Bottom;
    Parent := Self;
  end;
  SetLength(FButtonArray, aButtonList.Count);
  for i := 0 to High(FButtonArray) do begin
    FButtonArray[i] := TDlgIconCaptionButton.Create(Self, aButtonList.arr[i].caption, aButtonList.arr[i].mr);
    FButtonArray[i].Parent := FButtonPanel;
  end;
end;

destructor TQuestionDlg.Destroy;
begin
  FIconCaptionRec.Bitmap.Free;
  inherited Destroy;
end;

{ TButtonRecList }

function TButtonRecList.GetCount: Integer;
begin
  Exit(Length(arr));
end;

function TButtonRecList.AddButton(const aCaption: String; aMR: TModalResult; anIsCancel, anIsDefault: Boolean): Integer;
begin
  Result := Length(arr);
  SetLength(arr, Succ(Result));
  arr[Result].Init(aCaption, aMR, anIsCancel, anIsDefault);
end;

function TButtonRecList.HasCancel(out btnIndex: Integer): Boolean;
var
  i: Integer;
begin
  btnIndex := -1;
  for i := 0 to High(arr) do
    if arr[i].isCancel then begin
      btnIndex := i;
      Exit(True);
    end;
  Result := False;
end;

function TButtonRecList.HasDefault(out btnIndex: Integer): Boolean;
var
  i: Integer;
begin
  btnIndex := -1;
  for i := 0 to High(arr) do
    if arr[i].isDefault then begin
      btnIndex := i;
      Exit(True);
    end;
  Result := False;
end;

procedure TButtonRecList.Init;
begin
  SetLength(arr, 0);
end;

{ TButtonRec }

procedure TButtonRec.Init(const aCaption: String; aMR: TModalResult; anIsCancel, anIsDefault: Boolean);
begin
  caption := aCaption;
  mr := aMR;
  isCancel := anIsCancel;
  isDefault := anIsDefault;
end;

end.

